@@ -6,13 +6,16 @@
 #' @aliases summary.rlgt
 #' @importFrom stats median
 #' @export
-#' @S3method print rlgtfit
+#' @exportS3Method print rlgtfit
 #' @method print rlgtfit
 #' @rdname lgt
 print.rlgtfit <- function(x, ...) {
   if(!inherits(x, "rlgtfit")) stop("not a legitimate result of an Rlgt model")
   
-  if (x$use.regression) {
+  if (is.null(x$use.regression)) {
+    # Schmidt's model
+    print(lapply(x,median))
+  } else if (x$use.regression) {
     temp <- lapply(x$params,median)
     temp[['regCoef']] <- colMeans(x$params[['regCoef']])
     print(temp)
